#! /usr/bin/python3

################################################################################
# flexiWAN SD-WAN software - flexiEdge, flexiManage.
# For more information go to https://flexiwan.com
#
# Copyright (C) 2021  flexiWAN Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
################################################################################

import jwt
import datetime
import secrets

# This module provides an API for token-based authorization management.
# In current implementation, it is assumed that only one authorization
# token per username is permitted.

access_tokens = set()

key = "flexiwan"
expires_after_minutes = 30


def _generate_access_token(username):
    """Generates new token

    Args:
        username (string): user name to associate the token with

    Returns:
        string: generated jwt token
    """
    access_token = jwt.encode(
        {
            "user": username,
            "exp": datetime.datetime.utcnow()
            + datetime.timedelta(minutes=expires_after_minutes),
            "token": secrets.token_hex(),
        },
        key,
        algorithm="HS256",
    ).decode()
    return access_token


def assign_access_token(username):
    # add protection - no more than 20 users
    """Generates new token and adds to database

    Args:
        username (string): user name to assign the token to

    Returns:
        string: generated jwt token
    """
    access_token = _generate_access_token(username)
    access_tokens.add(access_token)
    return access_token


def validate_access_token(token):
    """Validates access token against existing collection of tokens. If expired, the token is removed

    Args:
        token (str): encoded jwt token

    Returns:
        boolean: True if token is valid, otherwise False
    """
    if not token:
        print("Token invalid (empty)")
        return False
    if not token in access_tokens:
        print("Unknown token")
        return False

    # token found, now check expiration
    if access_token_expired(token):
        access_tokens.remove(token)
        return False

    return True


def access_token_expired(access_token):
    """Checks whether specified access token has expired

    Args:
        token (str): access token

    Returns:
        Boolean: True if the token has expired, otherwise False
    """
    try:
        jwt.decode(access_token, key, algorithms=["HS256"])
    except jwt.ExpiredSignatureError:
        # Signature has expired
        print("Access token invalid (expired)", access_token)
        return True
    return False


def check_expired_access_tokens():
    """Removes expired access tokens from the database"""
    for token in list(access_tokens):
        if access_token_expired(token):
            access_tokens.remove(token)
