#! /usr/bin/python3

################################################################################
# flexiWAN SD-WAN software - flexiEdge, flexiManage.
# For more information go to https://flexiwan.com
#
# Copyright (C) 2021  flexiWAN Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
################################################################################

import json
import time
import subprocess

import cherrypy
from mediator import Mediator


def stream_shell_command(command):
    """Runs supplied shell command

    Args:
        command (string): command to run

    Yields:
        string: responses (line by line) from the command output
    """
    print("stream_shell_command", command)
    shell_command = command.split()
    process = subprocess.Popen(
        shell_command,
        stdout=subprocess.PIPE,
        universal_newlines=True,
        stderr=subprocess.PIPE,
    )
    while True:
        output = process.stdout.readline()
        output = output.strip()
        if len(output):
            yield (output)
        # Do something else
        return_code = process.poll()
        if return_code is not None:
            print("RETURN CODE", return_code)
            # Process has finished, read rest of the output
            for output in process.stdout.readlines():
                output = output.strip()
                if len(output):
                    yield (output)
            break
    res = {"code": return_code, "error": process.stderr.readlines()}
    yield (json.dumps(res))


def data(str):
    """Converts input string into event stream data

    Args:
        str (string): source string

    Returns:
        string: source string formatted as event stream data
    """
    make_data = bytes("id: {}\ndata:%s\n\n" % (str), "utf8")
    return make_data


def traceroute(server):
    """Executes traceroute command

    Args:
        server (string): server address (IP or hostname)

    Yields:
        string: traceroute responses
    """
    yield data("Trying to traceroute to %s ..." % server)
    # check whether the command is available on the device
    # taken from here: https://stackoverflow.com/questions/45999519/test-if-program-exists-in-bash-script-abbreviated-version
    check_traceroute_cmd = '[ -x "$(command -v traceroute)" ] && echo True || echo False'
    res = subprocess.check_output(check_traceroute_cmd, shell=True).decode("utf-8").strip()
    if res == "False":
        yield data(json.dumps({"code": 2, "error": "Traceroute command not available on the device"}))
        return
    cmd = "traceroute %s" % server
    print("Calling stream_shell_command", cmd)
    for ln in stream_shell_command(cmd):
        yield (data(ln))


def ping(server):
    """Executes ping command (10 count)

    Args:
        server (string): server address (IP or hostname)

    Yields:
        string: ping responses
    """
    yield (data("Trying to ping to %s ..." % server))
    cmd = "ping -c 10 %s" % server
    print("Calling stream_shell_command", cmd)
    for ln in stream_shell_command(cmd):
        yield (data(ln))


def continuous_ping(server):
    """Executes continuous ping

    Args:
        server (string): server address (IP or hostname)

    Yields:
        string: ping responses
    """
    running_attempts = []
    WND_SIZE = 10
    THRESHOLD = 6
    for _ in range(10):
        cmd = ["fping", "--timeout=10000", "-C", "1", "-R", server]
        ping = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        std_out = ping.stdout.read().decode().rstrip()
        std_err = ping.stderr.read().decode().rstrip()
        delay = std_err.split(":")[1].strip()
        running_attempts.append(True if std_out is not "" else False)
        successes = running_attempts.count(True)
        failures = min(len(running_attempts), WND_SIZE) - successes
        # Compose response. The last response will also include the connectivity status
        res = {
            "server": server,
            "result": "success" if std_out is not "" else "failure",
            "delayMs": delay if std_out is not "" else "",
            "error": "" if std_out is not "" else delay,
            "extended": std_err,
            "connected": failures < THRESHOLD
            if len(running_attempts) == WND_SIZE
            else None,
        }
        yield (data(json.dumps(res)))
        time.sleep(1)


@cherrypy.expose
class ConnectivityModule:
    def __init__(self):
        self.mediator = Mediator()

    @cherrypy.tools.accept(media="text/plain")
    @cherrypy.tools.json_out()
    @cherrypy.popargs("type")
    @cherrypy.popargs("server")
    def GET(self, type, server):
        if not server:
            cherrypy.response.status = 500
            yield from {"error": "Invalid input"}
        server = server.split(" ", 1)[0]
        if not server:
            cherrypy.response.status = 500
            yield from {"error": "Invalid input"}
        if type == "traceroute":
            yield from traceroute(server)
        if type == "ping":
            yield from ping(server)
        if type == "continuousping":
            print("Received continuousping request")
            yield from continuous_ping(server)
        cherrypy.response.status = 500
        return {"error": "Invalid input"}

    GET._cp_config = {"response.stream": True}
