#! /usr/bin/python3

################################################################################
# flexiWAN SD-WAN software - flexiEdge, flexiManage.
# For more information go to https://flexiwan.com
#
# Copyright (C) 2021  flexiWAN Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
################################################################################

import datetime
import json
import glob
import os
import subprocess
import sys
import yaml
import ruamel.yaml
import re
import uuid
import platform
import importlib
import ipaddress
import time

# Import agent files
agent_path = os.path.realpath(
    os.path.join(os.path.dirname(os.path.realpath(__file__)), "../../")
)
print("Agent path %s" % agent_path)
sys.path.append(agent_path)

# Import system checker files
system_checker_path = os.path.realpath(
    os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "../../tools/system_checker/"
    )
)
print("System checker path %s" % system_checker_path)
sys.path.append(system_checker_path)

import fwglobals
import fwutils
import fwsystem_checker as ch
import fwsystem_checker_common
import fwnetplan
import fwpppoe


def _get_netplan_dirs():
    return ["/etc/netplan/*.yaml", "/lib/netplan/*.yaml", "/run/netplan/*.yaml"]


def _get_device_token_filename():
    """Gets the name of the token file configured on this system

    Returns:
        string: token file name
    """
    fwagent_conf_file_path = fwglobals.g.FWAGENT_CONF_FILE
    print("fwagent_conf_file_path", fwagent_conf_file_path)
    with open(fwagent_conf_file_path, "r") as conf_file:
        conf = yaml.load(conf_file, Loader=yaml.SafeLoader)
    agent_conf = conf.get("agent", {})
    return agent_conf.get("token", {})

def _get_device_uuid():
    # Check if fwagent configuration file has the simulated uuid
    with open(fwglobals.g.FWAGENT_CONF_FILE, 'r') as stream:
        config = yaml.safe_load(stream)
        uuid = config.get('agent', []).get('uuid', None)
        if uuid:
            return uuid
    # If simulated uuid is not found, get the hardware uuid
    uuid_filename = '/sys/class/dmi/id/product_uuid'
    found_uuid = subprocess.check_output(['cat', uuid_filename]).decode().split('\n')[0].strip()
    return found_uuid if found_uuid else None

class Mediator(object):
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            print("Creating the Mediator object")
            cls._instance = super(Mediator, cls).__new__(cls)
            # Put any initialization here.
            fwglobals.initialize()
        return cls._instance

    def get_interfaces(self):
        try:
            return True, _get_interfaces()
        except subprocess.CalledProcessError as ex:
            err, desc = ex.args
            print("subprocess Exception happened %s, %s" % (err, desc))
            return False, "Failed to get device interfaces"
        except Exception as ex:
            err, desc = ex.args
            print("Exception happened %s, %s" % (err, desc))
            return False, err

    def get_device_info(self):
        router_state, router_reason = fwutils.get_router_status()
        print("router state %s, error %s" % (router_state, router_reason))
        device_management_status = _get_device_management_status()
        print("device management status %s" % (device_management_status))
        machine_id = _get_device_uuid()
        serial_number = fwutils.get_machine_serial()
        hostname = os.uname()[1]
        if not hostname:
            hostname = "unknown"
        versions = fwutils.get_device_versions(fwglobals.g.VERSIONS_FILE)
        return {
            "vRouter": {"state": router_state, "error": router_reason},
            "deviceManagement": device_management_status,
            "deviceSerialNumber": serial_number if serial_number != "0" else "N/A",
            "machineId": machine_id,
            "versions": versions,
            "hostname": hostname,
        }

    def get_device_token(self, get_token_file=_get_device_token_filename):
        """Attempts to read device token

        Args:
            get_token_file (function, optional): token file retrieval function. Defaults to _get_device_token_filename.

        Returns:
            tuple: result, token
        """
        token_file = get_token_file()
        return _get_device_token(token_file)

    def generate_uuid(self, machine_id):
        uuid = ""
        try:
            uuid = _generate_uuid(machine_id)
        except Exception as ex:
            print("Exception", type(ex))
            print(ex.args)
            return False, ex.args
        # reset the agent
        res = _run_shell_command("systemctl restart flexiwan-router")
        print("systemctl restart flexiwan-router", res)
        # since the service is restarted, need to reload the configuration
        # in order for the fwglobals to pick the updated uuid
        fwglobals.g.cfg = fwglobals.g.FwConfiguration(
            fwglobals.g.FWAGENT_CONF_FILE, fwglobals.g.DATA_PATH
        )
        return True, uuid

    def get_netplan_dirs(self):
        return _get_netplan_dirs()

    def get_all_netplan_ethernets(self, netplan_files):
        """processes netplan files collection and generates dictionary
        in the form of {"if_name", "path_to_file"}

        Args:
            netplan_files (list): list of netplan files to process

        Returns:
            dict: netplan ethernet interfaces and their container files \
            (e.g. {"enp0s3": "/etc/netplan/01-netcfg.yaml"})
        """
        ethernets = {}
        if not netplan_files:
            return ethernets

        for netplan_file in netplan_files:
            res = map(
                lambda x: {x: netplan_file},
                _get_netplan_file_ethernets(netplan_file),
            )
            for item in res:
                ethernets.update(item)
        print("Found the following ethernet interfaces:")
        print(json.dumps(ethernets))
        return ethernets

    def remove_interface(self, file_name, interface):
        config = None
        with open(file_name, "r") as stream:
            config = yaml.safe_load(stream)
            if not config:
                return False
            network = config["network"]
            ethernets = network["ethernets"]
            if interface in ethernets:
                ethernets.pop(interface)
            # don't leave ethernets section if no interfaces are left
            if len(ethernets) == 0:
                network.pop("ethernets")

        with open(file_name, "w") as stream:
            yaml.safe_dump(config, stream, default_flow_style=False)

    def add_interface(self, file_name, interface_name, interface_config):
        """Adds interface to an existing netplan file

        Args:
            file_name (string): netplan file
            interface_name (string): interface name
            interface_config (dict): interface configuration
        """
        config = None
        with open(file_name, "r") as stream:
            config = yaml.safe_load(stream)
            if not config:
                return False
            network = config["network"]
            if not "ethernets" in network:
                network["ethernets"] = {}
            ethernets = network["ethernets"]
            ethernets[interface_name] = interface_config

        with open(file_name, "w") as stream:
            yaml.safe_dump(config, stream, default_flow_style=False)

    def generate_netplan_configuration(self, interface):
        """generates netplan configuration from the specified interface

        Args:
            interface (dict): interface configuration. example:
            {
                "name": "enp0s3",
                "devId": "pci:0000:00:03.00",
                "driver": "e1000",
                "MAC": "08:00:27:2f:be:47",
                "IPv4": "192.168.1.69/24",
                "IPv4Mask": "24",
                "IPv6": "",
                "IPv6Mask": "",
                "dhcp": "no",
                "gateway": "192.168.1.1",
                "metric": "150",
                "internetAccess": true,
                "public_ip": "108.52.21.48",
                "public_port": 4789,
                "nat_type": "Full Cone"
            }

        Returns:
            [dict]: netplan compatible configuration.
            Example: '{"dhcp4": false, "addresses": ["192.168.1.69/24"], "routes": [{"to": "0.0.0.0/0", "via": "192.168.1.1", "metric": "150"}]}'
        """
        config_section = {}
        metric = interface.get("metric", "")
        metric = 0 if not metric else int(metric)

        try:
            if re.match("yes", interface["dhcp"]):
                config_section["dhcp4"] = True
                config_section["dhcp4-overrides"] = {"route-metric": metric}
                use_dns_servers = interface.get("useDhcpDnsServers")
                config_section["dhcp4-overrides"]["use-dns"] = (
                    True if use_dns_servers is None else interface["useDhcpDnsServers"]
                )
            else:
                config_section["dhcp4"] = False
                config_section["addresses"] = [interface["IPv4"]]

                gw = interface["gateway"]
                if gw:
                    routes = config_section.get("routes", [])
                    routes.append({"to": "0.0.0.0/0", "via": gw, "metric": metric})
                    config_section[
                        "routes"
                    ] = routes  # Handle case where there is no 'routes' section
            # set up dns
            addresses = interface.get("dnsServers", [])
            search = interface.get("dnsDomains", [])
            if len(addresses) > 0 or len(search) > 0:
                config_section["nameservers"] = {}
                nameservers = config_section["nameservers"]
                nameservers["addresses"] = addresses
                nameservers["search"] = search
            name = interface["name"]
            print(f"generate_netplan_configuration: Generated configuration for {name}")
            print(json.dumps(config_section))
            return config_section

        except Exception as ex:
            err, desc = ex.args
            print("Exception happened %s, %s" % (err, desc))
            return {}

    def get_next_available_netplan_file_name(self, location):
        """Looks up for the next available (unused) netplan file name
        in format xx-netcfg.baseline.yaml.\n
        The netplan update mechanism works in the following way:
        - if the interface to be updated is found in the initial netplan file (e.g. 01-netcfg.yaml),
        create a reserved copy of it with fworig extension (e.g. 01-netcfg.yaml.fworig). The .fworig file
        never gets modified. Then rename the initial file to a baseline file (e.g. 01-netcfg.baseline.yaml) and
        update it with the modified settings.
        - if the interface to be updated is not found in any of the netplan files, then create
        a new baseline file (e.g. 02-netcfg.baseline.yaml) and update it with the modified settings
        - if the interface to be updated is found in the baseline file (e.g. 03-netcfg.baseline.yaml),
        update it with the modified settings.

        Args:
            location (str): path to netplan directory

        Returns:
            [str]: netplan yaml file name
        """
        for i in range(1, 50):
            candidate_name = "%02d-netcfg.baseline.yaml" % i
            if not (
                os.path.exists(os.path.join(location, candidate_name))
                or os.path.exists(os.path.join(location, "%02d-netcfg.yaml" % i))
            ):
                return candidate_name
        return ""

    def update_device_interfaces(self, interface, get_netplan_dirs=_get_netplan_dirs):
        """Applies the changes in existing device interface

        Args:
            interface (object): interface object
            get_netplan_dirs (list, optional): list of netplan directories. Defaults to _get_netplan_dirs.

        Returns:
            boolean: operation status
            object: updated interface
        """
        # cannot update interfaces while router is not stopped
        router_state, router_reason = fwutils.get_router_status()
        print("router state %s, error %s" % (router_state, router_reason))
        if router_state != "stopped":
            return False, "Cannot update interfaces while router is not stopped"

        interface_name = interface["name"]
        print(f"Updating interface {interface_name}")
        existing_interfaces = _get_interfaces()

        res, error = check_interface_exists(interface_name, existing_interfaces)
        if not res:
            print(f"Interface {interface_name} does not exist, returning")
            return False, error

        # valid IPv4 address is required when interface type is static
        pppoe = interface.get("pppoe", {})
        pppoe_config = pppoe.get("config")
        pppoe_enabled = pppoe_config and pppoe_config["is_enabled"]
        if not interface["IPv4"] and interface["dhcp"] == "no" and not pppoe_enabled:
            return False, "Interface %s is missing a valid IPv4 address" % (
                interface_name
            )

        # Check address overlap only when the updated interface is static. There is no control
        # of the ip address in the case of dhcp interface, so this makes the ip overlap check pointless,
        # but only for when the interface is static.
        if interface["dhcp"] == "no":
            res, error = check_ip_address_overlap(interface, existing_interfaces)
            if not res:
                return False, error

        # Check for same metric values only between the wan interfaces.
        # In the case of pppoe being turned on, the interface doesn't have
        # the gateway available yet (until pppoe is actually enabled), but
        # still need to check for duplicate metric before enabling it.
        if len(interface.get("gateway", "")) or pppoe_enabled:
            # get the wan interfaces (the ones which have a gateway configured)
            wan_interfaces = list(
                filter(
                    lambda existing_interface: len(
                        existing_interface.get("gateway", "")
                    )
                    > 0 or existing_interface.get("pppoe", {"config": {"is_enabled": False}})["config"]["is_enabled"] is True,
                    existing_interfaces,
                )
            )
            print(
                f"update_device_interfaces: Found wan interfaces to check for same metric against {interface_name}"
            )
            print(wan_interfaces)
            res, error = _check_same_metric(interface, wan_interfaces)
            if not res:
                return False, error

        device_type = interface.get("deviceType", "dpdk")

        if device_type == "dpdk" or device_type == "pppoe":
            return self.update_dpdk_interface(interface, get_netplan_dirs)
        elif device_type == "lte":
            return self.update_lte_interface(interface)
        else:
            # wifi and others are not supported
            print("update_device_interfaces: unsupported device type", device_type)
            return False, f"Unsupported device type {device_type}"

    def _is_pppoe_config_equal(self, pppoe_conf, if_conf):
        if pppoe_conf and if_conf:
            return (
                pppoe_conf.is_enabled == if_conf["is_enabled"]
                and pppoe_conf.user == if_conf["username"]
                and pppoe_conf.password == if_conf["password"]
                and pppoe_conf.mtu == if_conf["mtu"]
                and pppoe_conf.mru == if_conf["mru"]
                and pppoe_conf.usepeerdns == if_conf["usepeerdns"]
                and set(pppoe_conf.nameservers) == set(if_conf["nameservers"])
            )
        if not pppoe_conf and not if_conf:
            return True
        return False

    def update_dpdk_interface(self, interface, get_netplan_dirs):
        # Perform initial validation steps
        print("Calling soft_check_multiple_interface_definitions")
        res = (
            fwsystem_checker_common.Checker().soft_check_multiple_interface_definitions()
        )
        print(
            "Call to soft_check_multiple_interface_definitions resulted with %s" % (res)
        )
        if not res:
            return False, "Multiple interface definitions detected, cannot continue"

        print("Calling soft_check_duplicate_netplan_sections")
        res = fwsystem_checker_common.Checker().soft_check_duplicate_netplan_sections()
        if not res:
            return False, "Duplicate netplan sections detected, cannot continue"
        print("Call to soft_check_duplicate_netplan_sections resulted with %s" % (res))

        interface_name = interface["name"]

        print("PPPoE: checking configuration")
        client = fwpppoe.FwPppoeClient(
            fwglobals.g.PPPOE_DB_FILE,
            fwglobals.g.PPPOE_CONFIG_PATH,
            fwglobals.g.PPPOE_CONFIG_PROVIDER_FILE,
        )
        if client is None:
            print("PPPoE: create FwPppoeClient instance failed")
            return True, interface
        dev_id = interface["devId"]
        print(f"Getting PPPoE config for {interface_name}, {dev_id}")
        pppoe_conf = client.get_interface(interface_name, dev_id)
        print(f"PPPoE: existing configuration for {interface_name}: {str(pppoe_conf)}")

        pppoe = interface.get("pppoe", {})
        pppoe_config = pppoe.get("config")
        metric = interface.get("metric", 0)
        # since metric is set outside the pppoe configuration, there is a chance that the users
        # just want to modify the metric of the existing pppoe configuration, so need to check this
        # condition as well
        if not self._is_pppoe_config_equal(pppoe_conf, pppoe_config) or (
            pppoe_conf is not None and pppoe_conf.metric != metric
        ):
            print(
                f"PPPoE: setting for {interface_name}, {dev_id}, {json.dumps(pppoe_config)}, metric {metric}"
            )

            device_management_status = _get_device_management_status()
            if device_management_status == "stopped":
                print(
                    "update_dpdk_interface: failed to configure pppoe interface, device management is not running"
                )
                return False, "Failed to configure pppoe interface, device management is not running"

            # remove pppoe interface first
            print(f"PPPoE: removing {interface_name}, {dev_id}")
            res = _run_shell_command(f'sudo fwagent configure pppoe interfaces delete --if_name {interface_name}')
            if res != "":
                print(
                    "update_dpdk_interface: failed to remove pppoe interface, error %s"
                    % (res)
                )
                return False, "Failed to remove pppoe interface"
            print(f"PPPoE: {interface_name}, {dev_id} removed")

            try:
                metric = int(metric) if len(metric) > 0 else 0
            except Exception as e:
                print(
                    "update_dpdk_interface: invalid metric value, exception %s"
                    % (str(e))
                )
                return False, "Invalid metric value specified"

            if pppoe_config["is_enabled"] is True:
                print(f"PPPoE: adding {interface_name}, {dev_id}")
                command = ("sudo fwagent configure pppoe interfaces create",
                    f'--if_name {interface_name}',
                    f'--enabled True',
                    f'--user {pppoe_config["username"]}',
                    f'--password {pppoe_config["password"]}',
                    f'--mtu {int(pppoe_config["mtu"])}',
                    f'--mru {int(pppoe_config["mru"])}',
                    f'--metric {int(metric)}',
                    f'--usepeerdns {pppoe_config["usepeerdns"]}')
                command = " ".join(command)
                nameservers = pppoe_config["nameservers"]
                if pppoe_config["usepeerdns"] == False and len(nameservers) > 0:
                    command = command + f' --nameserver1={nameservers[0]}'
                    if len(nameservers) > 1:
                        command = command + f' --nameserver2={nameservers[1]}'
                res = _run_shell_command(command)
                if res != "":
                    print(
                        "update_dpdk_interface: failed to add pppoe interface, error %s"
                        % (res)
                    )
                    return False, "Failed to add pppoe interface"

                time.sleep(5)
                print(f"PPPoE: {interface_name}, {dev_id} added")
            # do not continue configuration of the interface after pppoe is enabled
            return True, interface
        print(f"PPPoE: {interface_name}, {dev_id} config unchanged, nothing to do")

        # get netplan files from all specified netplan directories
        print("update_dpdk_interface: getting netplan directories")
        netplan_dirs = get_netplan_dirs()
        print("update_dpdk_interface: netplan directories:")
        print(json.dumps(netplan_dirs))
        netplan_files = _get_netplan_files(netplan_dirs)
        ethernets = _get_all_netplan_ethernets(netplan_files)
        # generate new configuration based on the interface
        config = self.generate_netplan_configuration(interface)
        netplan_file_name = ""
        if interface_name in ethernets.keys():
            netplan_file_name = ethernets[interface_name]
            print(
                "update_dpdk_interface: found %s interface in %s file"
                % (interface_name, netplan_file_name)
            )
            # if interface is found not in the baseline file, then make reserved copy
            # of the netplan file to be modified
            netplan_file_name = fwnetplan.create_baseline_if_not_exist(
                netplan_file_name
            )
            # remove interface first
            self.remove_interface(netplan_file_name, interface_name)
        else:
            print(
                "update_dpdk_interface: interface %s not found in any of the existing netplan files"
                % interface_name
            )
            main_netplan_dir = "/etc/netplan"
            netplan_file_name = self.get_next_available_netplan_file_name(
                main_netplan_dir
            )
            netplan_file_name = os.path.join(main_netplan_dir, netplan_file_name)

            # create new netplan file
            print("Creating new netplan file %s" % netplan_file_name)
            fwnetplan._add_netplan_file(netplan_file_name)

        # add modified interface
        self.add_interface(netplan_file_name, interface_name, config)

        # apply modified configuration
        fwutils.netplan_apply("edgeui_update_device_interfaces")

        return True, interface

    def update_lte_interface(self, interface):
        dev_name = interface["name"]
        dns_servers = interface["dnsServers"]
        dev_id = interface["devId"]
        print("update_lte_interface: device", dev_name, dev_id)
        nic_name = fwutils.dev_id_to_linux_if(dev_id)
        if not nic_name:
            return False, "Failed to update interface %s" % (dev_name)
        set_dns_str = " ".join(map(lambda server: "--set-dns=" + server, dns_servers))
        if set_dns_str:
            try:
                print(
                    "update_lte_interface: calling systemd-resolve",
                    set_dns_str,
                    nic_name,
                )
                os.system("systemd-resolve %s --interface %s" % (set_dns_str, nic_name))
                return True, interface
            except Exception as e:
                print(
                    "update_lte_interface: failed to update interface, exception %s"
                    % (str(e))
                )
                return False, "Failed to update interface %s" % (dev_name)
        print(
            "update_lte_interface: failed to generate set dns string, device id",
            dev_name,
        )
        return False, "Failed to update interface %s" % (dev_name)

    def update_device_tokens(self, tokens):
        """
        Updates device tokens
        """
        print("Received update_device_tokens")
        return _update_device_tokens(tokens)

    def apply_device_command(self, command):
        """
        docstring
        """
        print("Received apply_device_command")
        return _apply_device_command(command)

    def get_device_logs(self, filter, lines):
        return _get_device_logs({"filter": filter, "lines": lines})

    def dump(self):
        """
        Generates dump for download
        """
        hostname = os.uname()[1]
        if not hostname:
            hostname = "unknown-hostname"
        filename = "fwdump_%s_%s.tar.gz" % (
            hostname,
            datetime.datetime.now().strftime("%Y%m%d_%H%M%S"),
        )
        fwutils.dump(filename=filename, path=fwglobals.g.DUMP_FOLDER)
        # get the dump
        files = glob.glob("%s/%s" % (fwglobals.g.DUMP_FOLDER, filename))
        print(json.dumps(files))
        return os.path.join(fwglobals.g.DUMP_FOLDER, files[0])

    def change_password(self, username, password):
        """
        changes password for given username
        """
        res = _change_password(user_name=username, new_password=password)
        if res == 0:
            return True, ""
        return False, "Failed to modify password"

    def get_system_checkers(self):
        """
        gets the available system checkers
        """
        return True, _get_system_checkers()

    def run_system_checkers(self, auto_fix):
        """
        runs the available system checkers
        """
        return True, _run_system_checkers(auto_fix)

    def run_system_checker(self, checker_name, auto_fix):
        """
        runs specified system checker
        """
        return _run_system_checker(checker_name, auto_fix)

    def get_device_jobs(self):
        return _get_device_jobs()


def _change_password(user_name, new_password):
    print("user_name", user_name)
    # The assumption here is that the user which owns the web service process
    # has sudo privileges, otherwise the chpasswd command will fail.
    command = "chpasswd"
    print("command", command)
    command = command.split()

    # first command echoes password into the stdout
    # in order to be piped into the second command with actual
    # change password command.
    cmd1 = subprocess.Popen(
        ["echo", "{0}:{1}".format(user_name, new_password)],
        stdout=subprocess.PIPE,
    )
    cmd2 = subprocess.Popen(command, stdin=cmd1.stdout, stdout=subprocess.PIPE)

    result = cmd2.wait()
    print("Command completed", command)
    print("result", result)
    return result


def _run_shell_command(command):
    """
    Runs supplied shell command
    """
    print("_run_shell_command", command)
    shell_command = command.split()
    process = subprocess.Popen(
        shell_command, stdout=subprocess.PIPE, universal_newlines=True
    )
    out = ""
    while True:
        output = process.stdout.readline()
        print(output.strip())
        out = out + output.strip()
        # Do something else
        return_code = process.poll()
        if return_code is not None or return_code != 0:
            print("RETURN CODE", return_code)
            # Process has finished, read rest of the output
            for output in process.stdout.readlines():
                print(output.strip())
                out = out + output.strip()
            break
    return out


def _get_device_management_status():
    """
    Applies 'fwagent show --status daemon' shell command
    """
    res = subprocess.check_output("fwagent show --status daemon", shell=True)
    print("_get_device_management_status", res)
    if b"not running" in res:
        return "stopped"
    return res.decode("utf-8")


def _get_interfaces():
    """Gets available device interfaces using fwagent cli.

    Returns:
        list: collection of interfaces
    """
    command = "fwagent cli -f ./commands/get-device-info.cli"
    res = subprocess.check_output(command, shell=True).decode("utf-8")
    # The response from the cli has the following format
    #
    # {
    # "message":{
    #     "network":{
    #         "interfaces":[
    #             {
    #             "name":"enp0s3",
    #             ...
    #             }
    #         ]
    #     },
    # },
    # "ok":1,
    # "entity":"agentReply"
    # }
    #
    # and is captured between return-value-start and return-value-end
    start = res.find("return-value-start") + len("return-value-start") + 1
    end = res.find("return-value-end") - 1
    if start == -1 or end == -1 or start >= end:
        raise Exception("Failed to get device interfaces", "Invalid command response")
    cli_response = res[start:end]
    cli_deserialized = json.loads(cli_response)

    if "ok" not in cli_deserialized.keys():
        print("ok key not found")
        raise Exception("Failed to get device interfaces", "Invalid command response")
    ok = cli_deserialized["ok"]

    if "entity" not in cli_deserialized.keys():
        print("entity key not found")
        raise Exception("Failed to get device interfaces", "Invalid command response")
    entity = cli_deserialized["entity"]

    if ok != 1 or entity != "agentReply":
        print(
            f"Expected entity=agentReply and ok=1, but got entity={entity} and ok={ok}"
        )
        raise Exception("Failed to get device interfaces", "Invalid command response")

    if "message" not in cli_deserialized.keys():
        print("message key not found")
        raise Exception("Failed to get device interfaces", "Invalid command response")
    message = cli_deserialized["message"]

    if "network" not in message.keys():
        print("network key not found")
        raise Exception("Failed to get device interfaces", "Invalid command response")
    network = message["network"]

    if "interfaces" not in network.keys():
        print("interfaces key not found")
        raise Exception("Failed to get device interfaces", "Invalid command response")
    interfaces = network["interfaces"]
    print(interfaces)
    client = fwpppoe.FwPppoeClient(
        fwglobals.g.PPPOE_DB_FILE,
        fwglobals.g.PPPOE_CONFIG_PATH,
        fwglobals.g.PPPOE_CONFIG_PROVIDER_FILE,
    )
    for interface in interfaces:
        interface_name = interface["name"]
        command = "systemd-resolve --status " + interface_name
        res = subprocess.check_output(command, shell=True).decode("utf-8")
        print("DNS configuration for interface %s" % (interface_name))
        print(res)
        dns = _get_interface_dns(res, interface_name)
        interface["dnsServers"] = dns.get("DNS Servers", [])
        interface["dnsDomains"] = dns.get("DNS Domain", [])

        # assume dpdk if not present
        device_type = interface.get("deviceType", "dpdk")

        # WiFi is static, so no need to check for dhcp dns
        if device_type == "wifi":
            continue

        # TODO: need to figure out how to determine whether or not lte
        # TODO: device is configured to use DHCP to get the DNS config.
        if device_type == "lte":
            continue

        # check whether the configured DNS is actually set up in netplan
        # in order to determine whether or not it is custom configured one or set up by
        # DHCP. Not applicable for Static.
        if interface["dhcp"] == "yes":
            netplan_dirs = _get_netplan_dirs()
            netplan_files = _get_netplan_files(netplan_dirs)
            ethernets = _get_all_netplan_ethernets(netplan_files)
            if interface_name in ethernets.keys():
                interface["useDhcpDnsServers"] = _is_dns_from_dhcp(
                    interface_name, ethernets[interface_name]
                )
        # get PPPoE configuration
        dev_id = interface["devId"]
        print(f"Getting PPPoE config for {interface_name}, {dev_id}")
        pppoe_conf = client.get_interface(interface_name, dev_id)
        print(f"PPPoE for {interface_name}: {str(pppoe_conf)}")
        if pppoe_conf:
            pppoe = {
                "config": {
                    "is_enabled": pppoe_conf.is_enabled,
                    "username": pppoe_conf.user,
                    "password": pppoe_conf.password,
                    "mtu": str(pppoe_conf.mtu),
                    "mru": str(pppoe_conf.mru),
                    "usepeerdns": pppoe_conf.usepeerdns,
                    "nameservers": pppoe_conf.nameservers,
                    "metric": str(pppoe_conf.metric),
                },
                "status": {
                    "is_connected": pppoe_conf.is_connected,
                    "addr": pppoe_conf.addr,
                    "gw": pppoe_conf.gw,
                },
            }
            print(f"PPPoE for {interface_name} response: {str(pppoe)}")
            interface["pppoe"] = pppoe
    print("_get_interfaces response:")
    print(interfaces)
    return interfaces


def _is_dns_from_dhcp(interface_name, netplan_file_name):
    with open(netplan_file_name, "r") as stream:
        config = yaml.safe_load(stream)
        if not config:
            # something wrong with the netplan file, assume use ones from DHCP
            print("_is_dns_from_dhcp Failed to load file", netplan_file_name)
            return True
        network = config["network"]
        if "ethernets" not in network:
            print(
                "_is_dns_from_dhcp Ethernets section not found in file",
                netplan_file_name,
            )
            # ethernets not found in the netplan file, assume use ones from DHCP
            return True
        ethernets = network["ethernets"]
        if interface_name not in ethernets.keys():
            # interface name not found in the netplan file, assume use ones from DHCP
            print(
                "_is_dns_from_dhcp Interface not found in file",
                netplan_file_name,
                interface_name,
            )
            return True
        interface = ethernets[interface_name]
        overrides = interface.get("dhcp4-overrides", {})
        if not overrides:
            # no overrides - assume use ones from DHCP
            print(
                "_is_dns_from_dhcp dhcp4-overrides section not found in file",
                netplan_file_name,
                interface_name,
            )
            return True
        use_dns = overrides.get("use-dns", True)
        return use_dns


def _get_netplan_files(netplan_dirs):
    files = []
    for netplan_dir in netplan_dirs:
        files = files + glob.glob(netplan_dir)
    print("Netplan files")
    print(json.dumps(files))
    return files


def _get_netplan_file_ethernets(file_name):
    ethernets = {}
    with open(file_name, "r") as stream:
        config = yaml.safe_load(stream)
        if not config:
            return ethernets
        network = config["network"]
        if "ethernets" not in network:
            return ethernets
        ethernets = network["ethernets"]
    return ethernets.keys()


def _get_all_netplan_ethernets(netplan_files):
    """processes netplan files collection and generates dictionary
    in the form of {"if_name", "path_to_file"}

    Args:
        netplan_files (list): list of netplan files to process

    Returns:
        dict: netplan ethernet interfaces and their container files \
        (e.g. {"enp0s3": "/etc/netplan/01-netcfg.yaml"})
    """
    ethernets = {}
    if not netplan_files:
        return ethernets

    for netplan_file in netplan_files:
        res = map(
            lambda x: {x: netplan_file},
            _get_netplan_file_ethernets(netplan_file),
        )
        for item in res:
            ethernets.update(item)
    print("Found the following ethernet interfaces:")
    print(json.dumps(ethernets))
    return ethernets


def _get_device_token(token_file):
    """Attempts to read the token from specified token file

    Args:
        token_file (str): Token file to read from

    Returns:
        tuple: result, token
    """

    token = ""
    if token_file:
        print("_get_device_token: %s" % (token_file))
    else:
        err = "Token configuration not found"
        print(err)
        return True, {"deviceToken": token}
    try:
        with open(token_file, "r") as f:
            token = f.readline().rstrip()
            return True, {"deviceToken": token}
    except:
        user_err = "Failed to load token from %s" % (token_file)
        err = user_err + "%s (%s)" % (
            format(sys.exc_info()[1]),
            format(sys.exc_info()[0]),
        )
        print(err)
        return True, {"deviceToken": token}


def file_write_and_flush(f, data):
    """Wrapper over the f.write() method that flushes wrote content
    into the disk immediately

    :param f:       the python file object
    :param data:    the data to write into file
    """
    f.write(data)
    f.flush()
    os.fsync(f.fileno())


def _update_device_tokens(token):
    """
    Updates device token
    """
    print(json.dumps(token))
    token_file = _get_device_token_filename()
    print("Device token filename", token_file)

    if "deviceToken" not in token:
        return False, "Invalid input"

    try:
        with open(token_file, "w") as f:
            file_write_and_flush(f, token["deviceToken"])
            print("Token updated")
    except Exception as ex:
        print("Exception", type(ex))
        print(ex.args)
        return False, "Exception happened"
    return True, token


# TODO: combine into single send_cli_command
def _stop_router():
    """
    Applies stop router device command using cli
    """
    process = subprocess.Popen(
        ["fwagent", "cli", "-f", "./commands/stop-router.cli"],
        stdout=subprocess.PIPE,
        universal_newlines=True,
    )

    while True:
        output = process.stdout.readline()
        print(output.strip())
        # Do something else
        return_code = process.poll()
        if return_code is not None or return_code != 0:
            print("RETURN CODE", return_code)
            # Process has finished, read rest of the output
            for output in process.stdout.readlines():
                print(output.strip())
            break


def _apply_device_command(command):
    """
    Applies device command
    """
    print("_apply_device_command", command)
    if not "command" in command:
        return False, "Invalid input"
    try:
        success_response = {"status": "success", "description": ""}
        if command["command"] == "stopRouter":
            _stop_router()
            return True, success_response
        elif command["command"] == "startDeviceManagement":
            device_management_status = _get_device_management_status()
            if device_management_status == "running":
                return True, {
                    "status": "success",
                    "description": "Device management already running",
                }
            _run_shell_command("systemctl start flexiwan-router")
            return True, success_response
        elif command["command"] == "stopDeviceManagement":
            device_management_status = _get_device_management_status()
            if device_management_status == "stopped":
                return True, {
                    "status": "success",
                    "description": "Device management already stopped",
                }
            _run_shell_command("systemctl stop flexiwan-router")
            return True, success_response
        elif command["command"] == "clearDeviceConfiguration":
            router_state, router_reason = fwutils.get_router_status()
            print("router state %s, error %s" % (router_state, router_reason))
            if router_state == "running":
                print("Stopping router")
                _stop_router()
            else:
                print("Skipping stop router, router stopped")
            _run_shell_command("fwagent reset -s")
            return True, success_response
        elif command["command"] == "reset":
            router_state, router_reason = fwutils.get_router_status()
            print("router state %s, error %s" % (router_state, router_reason))
            if router_state == "running":
                print("Stopping router")
                _stop_router()
            else:
                print("Skipping stop router, router stopped")
            _run_shell_command("fwagent reset -q")
            return True, success_response
        elif command["command"] == "completeReset":
            router_state, router_reason = fwutils.get_router_status()
            print("router state %s, error %s" % (router_state, router_reason))
            if router_state == "running":
                print("Stopping router")
                _stop_router()
            else:
                print("Skipping stop router, router stopped")
            _run_shell_command("fwagent reset -p -q")
            return True, success_response
        elif command["command"] == "reboot":
            os.system("sleep 10 ; reboot")
            return True, success_response
        else:
            return False, "Command not recognized"
    except Exception as ex:
        print("Exception", type(ex))
        print(ex.args)
        return False, ex.args


def _get_device_logs(params):
    """Get device logs.
    :param params: Parameters from flexiManage: {"filter": fwagent, "lines": 100}
    :returns: Dictionary with logs and status code.
    """
    dl_map = {
        "fwagent": fwglobals.g.ROUTER_LOG_FILE,
        "syslog": fwglobals.g.SYSLOG_FILE,
        "dhcp": fwglobals.g.DHCP_LOG_FILE,
        "vpp": fwglobals.g.VPP_LOG_FILE,
        "ospf": fwglobals.g.FRR_LOG_FILE,
        "hostapd": fwglobals.g.HOSTAPD_LOG_FILE,
        "application_ids": fwglobals.g.APPLICATION_IDS_LOG_FILE,
        "resolv_conf": "/etc/resolv.conf",
    }
    file = dl_map.get(params["filter"], "")
    if file == "":
        return (
            False,
            "specified log type %s not found, valid types are fwagent, syslog, dhcp, vpp, ospf, hostapd, application_ids, resolv_conf"
            % params["filter"],
        )
    lines = 0
    try:
        lines = int(params["lines"])
    except:
        return False, "invalid lines value specified, valid range is between 1 and 1000"
    if (lines < 1) or (lines > 1000):
        return False, "invalid number of lines, valid range is between 1 and 1000"
    try:
        logs = fwutils.get_device_logs(file, lines)
        return True, {"logs": logs}
    except:
        return False, "Failed to get %s device logs" % params["filter"]


def _generate_uuid(machine_id):
    """Uses provided one or auto-generates new uuid and writes it to a conf file

    Args:
        machine_id (string): machine uuid

    Returns:
        string: updated machine uuid
    """
    print("Machine Id", machine_id)
    if len(machine_id) == 0:
        print("Auto-generating new uuid")
        new_uuid = str(uuid.uuid1()).upper()
    else:
        new_uuid = machine_id["machineId"]
        print("Using provided uuid", new_uuid)
    conf_file_name = fwglobals.g.FWAGENT_CONF_FILE
    print("conf_file_name", conf_file_name)
    f = open(conf_file_name, "r")
    ruamel_yaml = ruamel.yaml.YAML()
    conf = ruamel_yaml.load(f)
    conf["agent"]["uuid"] = new_uuid
    f.close()
    f = open(conf_file_name, "w")
    ruamel_yaml.dump(conf, f)
    f.close()
    return new_uuid


def _get_checker_module():
    """Gets the checker module (essentially, the fwsystem_checker_common.py)

    Returns:
        checker module
    """
    (flavor, version, _) = platform.linux_distribution()
    module_name = (flavor + version.replace(".", "")).lower()
    module = importlib.import_module(module_name)
    return module.Checker(False)


def _get_single_hard_checker(checker):
    (name, params) = list(checker.items())[0]
    category = "hard"
    severity = params[1]
    desc = params[2]
    return {
        "name": name,
        "description": desc,
        "category": category,
        "severity": severity,
        "result": None,
    }


def _run_single_hard_checker(checker):
    (name, params) = list(checker.items())[0]
    category = "hard"
    severity = params[1]
    desc = params[2]
    result = None
    reason = None

    with _get_checker_module() as obj:
        checker_func = getattr(obj, name)
        args = params[0]
        try:
            result = "Passed" if checker_func(args) == True else "Failed"
        except Exception as e:
            result = "Skipped"
            reason = str(e)

    return {
        "name": name,
        "description": desc,
        "category": category,
        "severity": severity,
        "result": result,
        "reason": reason,
    }


def _get_single_soft_checker(checker):
    category = "soft"
    (name, params) = list(checker.items())[0]
    desc = ch.checker_name_to_description(name)
    severity = params["severity"]
    return {
        "name": name,
        "description": desc,
        "category": category,
        "severity": severity,
        "result": None,
    }


def _run_single_soft_checker(checker):
    category = "soft"
    (name, params) = list(checker.items())[0]
    desc = ch.checker_name_to_description(name)
    severity = params["severity"]
    result = None
    reason = None

    with _get_checker_module() as obj:
        checker_func = getattr(obj, name)
        try:
            result = "Passed" if checker_func(fix=False, prompt=desc) else "Failed"
        except Exception as e:
            result = "Skipped"
            reason = str(e)
    return {
        "name": name,
        "description": desc,
        "category": category,
        "severity": severity,
        "result": result,
        "reason": reason,
    }


def _get_system_checkers():
    result_hard = map(_get_single_hard_checker, ch.hard_checkers)
    result_soft = map(_get_single_soft_checker, ch.soft_checkers)

    result = list(result_hard) + list(result_soft)
    print(list(result))
    return list(result)


def _run_system_checkers(auto_fix):
    """Runs all system checkers at once

    Args:
        auto_fix (boolean): currently not in use, assumed no auto fix is applied

    Returns:
        [list]:system checkers with list with their description and results
    """
    result_hard = map(_run_single_hard_checker, ch.hard_checkers)
    result_soft = map(_run_single_soft_checker, ch.soft_checkers)

    result = list(result_hard) + list(result_soft)
    print(list(result))
    return list(result)


def _run_system_checker(checker_name, auto_fix):
    """Runs single system checker

    Args:
        checker_name (string): checker name (e.g. hard_check_ram)
        auto_fix (boolean): when set to True, will attempt to run and silently fix the checker

    Returns:
        [tuple]: status, result
        status: True if success, otherwise False
        result: operation result if success, otherwise error description
    """
    # do not run system checker when the router is running
    router_state = fwutils.get_router_status()[0]
    if router_state == "running":
        print("error: cannot run fwsystem_checker when the router is running")
        return (
            False,
            "cannot run fwsystem_checker when the router is running, please stop router first",
        )

    if "hard" in checker_name:
        return _run_hard_system_checker(checker_name)
    if "soft" in checker_name:
        return _run_soft_system_checker(checker_name, auto_fix)
    return False, "Unknown checker %s" % checker_name


def _run_hard_system_checker(checker_name):
    """Runs single hard system checker

    Args:
        checker_name (string): checker name (e.g. hard_check_ram)

    Returns:
        [tuple]: status, result
        status: True if success, otherwise False
        result: operation result if success, otherwise error description
    """

    def checker_by_name(checker):
        (name, params) = list(checker.items())[0]
        return name == checker_name

    result = list(filter(checker_by_name, ch.hard_checkers))
    if len(result) == 0:
        return False, "Unknown checker %s" % checker_name

    checker = result[0]
    (name, params) = list(checker.items())[0]
    category = "hard"
    severity = params[1]
    desc = params[2]
    result = None
    reason = None

    with _get_checker_module() as obj:
        checker_func = getattr(obj, name)
        args = params[0]
        try:
            result = "Passed" if checker_func(args) == True else "Failed"
        except Exception as e:
            result = "Skipped"
            reason = str(e)

    return True, {
        "name": name,
        "description": desc,
        "category": category,
        "severity": severity,
        "result": result,
        "reason": reason,
    }


def _run_soft_system_checker(checker_name, auto_fix):
    """Runs single soft system checker

    Args:
        checker_name (string): checker name (e.g. hard_check_ram)
        auto_fix (boolean): when set to True, will attempt to run and silently fix the checker

    Returns:
        [tuple]: status, result
        status: True if success, otherwise False
        result: operation result if success, otherwise error description
    """

    def checker_by_name(checker):
        (name, params) = list(checker.items())[0]
        return name == checker_name

    result = list(filter(checker_by_name, ch.soft_checkers))
    if len(result) == 0:
        return False, "Unknown checker %s" % checker_name

    checker = result[0]
    (name, params) = list(checker.items())[0]
    category = "soft"
    desc = ch.checker_name_to_description(name)
    severity = params["severity"]
    interactive = "" if not "interactive" in params else params["interactive"]
    print("%s checker interactive setting is %s" % (desc, interactive))
    result = None
    reason = None

    with _get_checker_module() as obj:
        checker_func = getattr(obj, name)
        try:
            if checker_func(fix=False, prompt=desc):
                # Do not fix if check is true
                return True, {
                    "name": name,
                    "description": desc,
                    "category": category,
                    "severity": severity,
                    "result": "Passed",
                }
            if "must" in interactive:
                # Cannot fix if interactive is must
                return True, {
                    "name": name,
                    "description": desc,
                    "category": category,
                    "severity": severity,
                    "result": "Failed",
                }
            result = (
                "Passed"
                if checker_func(fix=auto_fix, silently=True, prompt=desc)
                else "Failed"
            )
        except Exception as e:
            result = "Skipped"
            reason = str(e)

    return True, {
        "name": name,
        "description": desc,
        "category": category,
        "severity": severity,
        "result": result,
        "reason": reason,
    }


def _get_interface_dns(data, interface_name):
    """Parses dns data coming in as an output of systemd-resolve --status <interface_name> command

    Args:
        data (str): systemd-resolve --status <interface_name> command output
        interface_name (str ): interface name

    Returns:
        dict: object which contains parsed data. For example: given the results of
        systemd-resolve --status enp0s3 command as input:

        Link 12 (enp0s3)
              Current Scopes: DNS
               LLMNR setting: yes
        MulticastDNS setting: no
              DNSSEC setting: no
            DNSSEC supported: no
                 DNS Servers: 192.168.1.1
                  DNS Domain: myfiosgateway.com

        the return value will look like this:

       {
           'Current Scopes': 'DNS',
           'LLMNR setting': 'yes',
           'MulticastDNS setting': 'no',
           'DNSSEC setting': 'no',
           'DNSSEC supported': 'no',
           'DNS Servers': ['192.168.1.1'],
           'DNS Domain': ['myfiosgateway.com']
        }

        An empty dict is returned when the result does not belong to the requested interface
    """
    res = {}
    print("_get_interface_dns: processing %s" % (interface_name))
    if data is None or not data:
        print("_get_interface_dns: empty data received, returning")
        return {}
    print(data)
    lines = data.splitlines()
    if not interface_name in lines[0]:
        return {}
    # Holds the field currently being processed. Used in multi-line cases.
    current_field = ""
    for line in lines[1:]:
        current_line = str.strip(line)
        if current_line.startswith(
            (
                "Current Scopes",
                "LLMNR setting",
                "MulticastDNS setting",
                "DNSSEC setting",
                "DNSSEC supported",
            )
        ):
            items = list(map(str.strip, current_line.split(":", 1)))
            current_field = items[0]
            # make it a single value
            res[current_field] = items[1]
        elif current_line.startswith(("DNS Servers", "DNS Domain")):
            items = list(map(str.strip, current_line.split(":", 1)))
            current_field = items[0]
            # make it a list
            res[current_field] = [items[1]]
        else:
            # multi-line field, just append to already parsed value
            res[current_field].append(current_line)
    print("_get_interface_dns: returning response for %s" % (interface_name))
    print(res)
    return dict(res)


def check_interface_exists(interface_name, existing_interfaces):
    """[summary]

    Args:
        interface_name (string): interface name
        existing_interfaces (list): existing interfaces

    Returns:
        boolean: operation status, True when success, otherwise False
        string: error description, empty check was successful
    """
    # check whether the interface is part of the existing interfaces
    existing_interfaces_names = list(map(lambda x: x["name"], existing_interfaces))
    print("check_interface_exists: available interfaces", existing_interfaces_names)
    if interface_name not in existing_interfaces_names:
        error = "Interface %s is not available on the device" % (interface_name)
        print(error)
        return False, error

    print(
        "check_interface_exists: interface %s is available on the device"
        % (interface_name)
    )
    return True, ""


def check_ip_address_overlap(interface, existing_interfaces):
    """Checks for IPv4 address overlap between the configured interface vs existing ones

    Args:
        interface (object): interface to be configured
        existing_interfaces (list): existing interfaces

    Returns:
        boolean: operation status, True when success, otherwise False
        string: error description, empty check was successful
    """
    interface_name = interface["name"]
    # the interface to configure is already a combination of IPv4 and a mask
    new_addr = interface["IPv4"]
    if not new_addr:
        print(
            "check_ip_address_overlap: no IPv4 address provided for interface %s"
            % (interface_name)
        )
        return True, ""
    for existing_interface in existing_interfaces:
        if existing_interface["IPv4"] and existing_interface["name"] != interface_name:
            # the existing interface needs to be combined into IPv4 and a mask
            available_addr = "{0}/{1}".format(
                existing_interface["IPv4"], existing_interface["IPv4Mask"]
            )
            print(
                "check_ip_address_overlap: checking overlap of addresses %s and %s"
                % (new_addr, available_addr)
            )
            # compose an IPv4Address object and check if configured interface overlaps
            # with any of the existing ones
            ipaddr1 = ipaddress.IPv4Network(available_addr, False)
            ipaddr2 = ipaddress.IPv4Network(new_addr, False)
            if ipaddr1.overlaps(ipaddr2):
                print(
                    "check_ip_address_overlap: interfaces %s and %s have an IP address overlap"
                    % (interface_name, existing_interface["name"])
                )
                return (
                    False,
                    "IP addresses of the configured interfaces have an overlap",
                )
    print(
        "check_ip_address_overlap: interface %s does not have an overlap with any of the existing interfaces"
        % (interface_name)
    )
    return True, ""


def _check_same_metric(interface, existing_interfaces):
    """Checks for same metric value between the configured interface vs existing ones

    Args:
        interface (object): interface to be configured
        existing_interfaces (list): existing interfaces

    Returns:
        boolean: operation status, True when success, otherwise False
        string: error description, empty check was successful
    """
    interface_name = interface["name"]
    for existing_interface in existing_interfaces:
        if existing_interface["name"] != interface_name:
            print(
                "Checking metrics of addresses %s and %s"
                % (interface_name, existing_interface["name"])
            )
            if interface["metric"] == existing_interface["metric"]:
                print(
                    "Interfaces %s and %s have the same metric value"
                    % (interface_name, existing_interface["name"])
                )
                return (
                    False,
                    "Duplicated metrics are not allowed on interfaces",
                )
    print(
        "Interface %s does not have the same metric value with any of the existing interfaces"
        % (interface_name)
    )
    return True, ""

def _get_device_jobs():
    """Get the requested device jobs.
    :returns: Dictionary with jobs and status code.
    """

    """Gets available device interfaces using fwagent cli.

    Returns:
        list: collection of interfaces
    """
    command = "fwagent show --database jobs"
    res = subprocess.check_output(command, shell=True).decode("utf-8")
    # The response from the cli is expected to have the following format
    #
    # [
    #   {
    #       "errors": [],
    #       "job_id": "5401",
    #       "ok": "",
    #       "received_at": "2022-05-30 17:06:29.946450",
    #       "request": "sync-device"
    #   },
    #   {
    #      "errors": [
    #        {
    #          "command": {
    #            "error": "add_remove_route({'addr': '48.2.0.0/16', 'via': '192.168.1.1', 'metric': 2, 'remove': True, 'dev_id': 'pci:0000:00:03.00'}): add_remove_route: dev_id_to_linux_if_name failed",
    #            "name": "add_remove_route"
    #          },
    #        "request": "remove-route"
    #        }
    #      ],
    #      "job_id": "5402",
    #      "ok": 1,
    #      "received_at": "2022-05-30 17:07:07.752519",
    #      "request": "sync-device"
    #    }
    # ]
    #

    try:
        jobs = json.loads(res)
        return True, {"jobs": jobs}
    except:
        return False, "Failed to get device jobs"
