import json
import os
import sys
import pytest

backend_path = os.path.realpath(
    os.path.join(os.path.dirname(os.path.realpath(__file__)), "../")
)
print("Backend path %s" % backend_path)
sys.path.append(backend_path)

from mediator import _get_interface_dns


class TestDns:
    @pytest.mark.parametrize(
        "dns_file,interface,expected",
        [
            (
                os.path.join(
                    os.path.dirname(os.path.realpath(__file__)),
                    "input/dns/dns_multiple_entries_servers_domain.txt",
                ),
                "enp0s3",
                {},
            ),
            (
                os.path.join(
                    os.path.dirname(os.path.realpath(__file__)),
                    "input/dns/dns_multiple_entries_servers_domain.txt",
                ),
                "vpp0",
                {
                    "Current Scopes": "DNS",
                    "LLMNR setting": "yes",
                    "MulticastDNS setting": "no",
                    "DNSSEC setting": "no",
                    "DNSSEC supported": "no",
                    "DNS Servers": ["8.8.8.8", "8.8.4.4"],
                    "DNS Domain": ["google.com", "yahoo.com"],
                },
            ),
            (
                os.path.join(
                    os.path.dirname(os.path.realpath(__file__)),
                    "input/dns/dns_multiple_entries_servers.txt",
                ),
                "vpp0",
                {
                    "Current Scopes": "DNS",
                    "LLMNR setting": "yes",
                    "MulticastDNS setting": "no",
                    "DNSSEC setting": "no",
                    "DNSSEC supported": "no",
                    "DNS Servers": ["8.8.8.8", "8.8.4.4"],
                    "DNS Domain": ["google.com"],
                },
            ),
            (
                os.path.join(
                    os.path.dirname(os.path.realpath(__file__)),
                    "input/dns/dns_multiple_entries_domain.txt",
                ),
                "vpp0",
                {
                    "Current Scopes": "DNS",
                    "LLMNR setting": "yes",
                    "MulticastDNS setting": "no",
                    "DNSSEC setting": "no",
                    "DNSSEC supported": "no",
                    "DNS Servers": ["8.8.8.8"],
                    "DNS Domain": ["google.com", "yahoo.com"],
                },
            ),
            (
                os.path.join(
                    os.path.dirname(os.path.realpath(__file__)),
                    "input/dns/dns_no_servers_no_domain.txt",
                ),
                "enp0s9",
                {
                    "Current Scopes": "none",
                    "LLMNR setting": "yes",
                    "MulticastDNS setting": "no",
                    "DNSSEC setting": "no",
                    "DNSSEC supported": "no",
                },
            ),
            (
                os.path.join(
                    os.path.dirname(os.path.realpath(__file__)),
                    "input/dns/dns_no_servers_no_domain.txt",
                ),
                "vpp0",
                {},
            ),
            (
                os.path.join(
                    os.path.dirname(os.path.realpath(__file__)),
                    "input/dns/dns_multiple_entries_servers_ipv6_first.txt",
                ),
                "eth1",
                {
                    "Current Scopes": "DNS",
                    "LLMNR setting": "yes",
                    "MulticastDNS setting": "no",
                    "DNSSEC setting": "no",
                    "DNSSEC supported": "no",
                    "DNS Servers": ["fe80::1213:31ff:fec9:4a50", "192.168.2.1"],
                    "DNS Domain": ["lan"],
                },
            ),
            (
                os.path.join(
                    os.path.dirname(os.path.realpath(__file__)),
                    "input/dns/dns_multiple_entries_servers_ipv6_second.txt",
                ),
                "eth1",
                {
                    "Current Scopes": "DNS",
                    "LLMNR setting": "yes",
                    "MulticastDNS setting": "no",
                    "DNSSEC setting": "no",
                    "DNSSEC supported": "no",
                    "DNS Servers": ["192.168.2.1", "fe80::1213:31ff:fec9:4a50"],
                    "DNS Domain": ["lan"],
                },
            ),
            (
                os.path.join(
                    os.path.dirname(os.path.realpath(__file__)),
                    "input/dns/dns_multiple_entries_servers_ipv6_only.txt",
                ),
                "eth1",
                {
                    "Current Scopes": "DNS",
                    "LLMNR setting": "yes",
                    "MulticastDNS setting": "no",
                    "DNSSEC setting": "no",
                    "DNSSEC supported": "no",
                    "DNS Servers": ["2606:4700:4700::1111", "2606:4700:4700::1001"],
                    "DNS Domain": ["lan"],
                },
            ),
            (
                os.path.join(
                    os.path.dirname(os.path.realpath(__file__)),
                    "input/dns/dns_multiple_entries_servers_no_domain.txt",
                ),
                "wlan0",
                {
                    "Current Scopes": "none",
                    "LLMNR setting": "yes",
                    "MulticastDNS setting": "no",
                    "DNSSEC setting": "no",
                    "DNSSEC supported": "no",
                    "DNS Servers": ["8.8.8.8", "8.8.4.4", "1.1.1.1"],
                },
            ),
            (
                os.path.join(
                    os.path.dirname(os.path.realpath(__file__)),
                    "input/dns/dns_multiple_current_scopes.txt",
                ),
                "wlp4s0",
                {
                    "Current Scopes": "DNS LLMNR/IPv4 LLMNR/IPv6",
                    "LLMNR setting": "yes",
                    "MulticastDNS setting": "no",
                    "DNSSEC setting": "no",
                    "DNSSEC supported": "no",
                },
            ),
        ],
    )
    def test_get_dnd_success_different_inputs(self, dns_file, interface, expected):

        data = ""
        with open(dns_file, "r") as f:
            data = f.read()

        response = _get_interface_dns(data, interface)
        assert response == expected

    @pytest.mark.parametrize(
        "data,expected",
        [
            (
                None,
                {},
            ),
            (
                "",
                {},
            ),
        ],
    )
    def test_get_dnd_success_empty(self, data, expected):

        response = _get_interface_dns(data, "whatever")
        assert response == expected
