import json
import os
import sys
import collections
import pytest

backend_path = os.path.realpath(
    os.path.join(os.path.dirname(os.path.realpath(__file__)), "../")
)
print("Backend path %s" % backend_path)
sys.path.append(backend_path)

# input_path = os.path.realpath(
#     os.path.join(os.path.dirname(os.path.realpath(__file__)), "input/")
# )
# print("Input path %s" % input_path)
# sys.path.append(input_path)

from mediator import Mediator
from mediator import _is_dns_from_dhcp
from mediator import _get_netplan_file_ethernets
from mediator import _get_netplan_files


def _get_netplan_dirs():
    current_dir = os.path.dirname(os.path.realpath(__file__))
    return [os.path.join(current_dir, "output/*.yaml")]


class TestNetplan:
    def test_get_netplan_dirs_success(self):
        expected = ["/etc/netplan/*.yaml", "/lib/netplan/*.yaml", "/run/netplan/*.yaml"]
        mediator = Mediator()
        netplan_dirs = mediator.get_netplan_dirs()
        assert collections.Counter(netplan_dirs) == collections.Counter(expected)

    @pytest.mark.parametrize(
        "file_name,expected",
        [
            ("input/empty.yaml", 0),
            ("input/01-netcfg.yaml", 1),
            ("input/02-netcfg.yaml", 2),
        ],
    )
    def test_get_netplan_file_ethernets_success(self, file_name, expected):
        mediator = Mediator()
        file_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), file_name)
        ethernets = _get_netplan_file_ethernets(file_path)
        assert len(ethernets) == expected

    def my_setup(self):
        # cleanup output folder
        cmd = "rm -r output"
        os.system(cmd)
        cmd = "mkdir output"
        os.system(cmd)

    def test_remove_interface_success(self):
        self.my_setup()

        current_dir = os.path.dirname(os.path.realpath(__file__))
        file_path = os.path.join(current_dir, "input/01-netcfg.yaml")
        output = os.path.join(current_dir, "output")
        cmd = "cp %s %s" % (file_path, output)
        os.system(cmd)

        work_file_path = os.path.join(current_dir, "output/01-netcfg.yaml")
        mediator = Mediator()
        interface_to_remove = "enp0s3"

        # load file and check interfaces
        ethernets = _get_netplan_file_ethernets(work_file_path)
        assert len(ethernets) == 1
        assert interface_to_remove in ethernets

        mediator.remove_interface(work_file_path, interface_to_remove)

        # load file again and check interfaces
        ethernets = _get_netplan_file_ethernets(work_file_path)
        assert len(ethernets) == 0
        assert interface_to_remove not in ethernets

    def test_remove_interface_not_existing_interface_success(self):
        self.my_setup()

        current_dir = os.path.dirname(os.path.realpath(__file__))
        file_path = os.path.join(current_dir, "input/01-netcfg.yaml")
        output = os.path.join(current_dir, "output")
        cmd = "cp %s %s" % (file_path, output)
        os.system(cmd)

        work_file_path = os.path.join(current_dir, "output/01-netcfg.yaml")
        mediator = Mediator()
        interface_to_remove = "enp0s8"

        # load file and check interfaces
        ethernets = _get_netplan_file_ethernets(work_file_path)
        assert len(ethernets) == 1

        mediator.remove_interface(work_file_path, interface_to_remove)

        # load file again and check interfaces
        ethernets = _get_netplan_file_ethernets(work_file_path)
        assert len(ethernets) == 1
        assert interface_to_remove not in ethernets

    def test_add_interface_success(self):
        self.my_setup()

        current_dir = os.path.dirname(os.path.realpath(__file__))
        file_path = os.path.join(current_dir, "input/01-netcfg.yaml")
        output = os.path.join(current_dir, "output")
        cmd = "cp %s %s" % (file_path, output)
        os.system(cmd)

        work_file_path = os.path.join(current_dir, "output/01-netcfg.yaml")
        mediator = Mediator()
        interface_name = "enp0s8"
        interface_config = {"dhcp4": True}

        # load file and check interfaces
        ethernets = _get_netplan_file_ethernets(work_file_path)
        assert len(ethernets) == 1
        assert interface_name not in ethernets

        mediator.add_interface(work_file_path, interface_name, interface_config)

        # load file again and check interfaces
        ethernets = _get_netplan_file_ethernets(work_file_path)
        assert len(ethernets) == 2
        assert interface_name in ethernets

    def test_get_netplan_files_success(self):
        self.my_setup()

        current_dir = os.path.dirname(os.path.realpath(__file__))
        input_dir = os.path.join(current_dir, "input/*.yaml")

        netplan_files = _get_netplan_files([input_dir])

        expected = [
            os.path.join(current_dir, "input/01-netcfg.yaml"),
            os.path.join(current_dir, "input/02-netcfg.yaml"),
            os.path.join(current_dir, "input/03-netcfg.yaml"),
            os.path.join(current_dir, "input/empty.yaml"),
            # os.path.join(current_dir, "input/dhcp-overrides.yaml"),
        ]
        assert collections.Counter(netplan_files) == collections.Counter(expected)

    def test_get_all_netplan_ethernets_success(self):
        self.my_setup()

        current_dir = os.path.dirname(os.path.realpath(__file__))
        mediator = Mediator()

        netplan_files = [
            os.path.join(current_dir, "input/02-netcfg.yaml"),
            os.path.join(current_dir, "input/03-netcfg.yaml"),
            os.path.join(current_dir, "input/empty.yaml"),
        ]

        ethernets = mediator.get_all_netplan_ethernets(netplan_files)

        expected = {
            "enp0s3": os.path.join(current_dir, "input/02-netcfg.yaml"),
            "enp0s8": os.path.join(current_dir, "input/02-netcfg.yaml"),
            "enp0s10": os.path.join(current_dir, "input/03-netcfg.yaml"),
        }

        print("get_all_netplan_ethernets returned:")
        print(ethernets)

        assert collections.Counter(ethernets) == collections.Counter(expected)

    @pytest.mark.parametrize(
        "interface,expected",
        [
            # (
            #     {
            #         "name": "enp0s3",
            #         "IPv4Mask": "24",
            #         "dhcp": "yes",
            #         "gateway": "192.168.1.1",
            #         "metric": "150",
            #     },
            #     '{"dhcp4": true, "dhcp4-overrides": {"route-metric": 150}}',
            # ),
            (
                {
                    "name": "enp0s3",
                    "devId": "pci:0000:00:03.00",
                    "driver": "e1000",
                    "MAC": "08:00:27:2f:be:47",
                    "IPv4": "192.168.1.69/24",
                    "IPv4Mask": "24",
                    "dhcp": "no",
                    "gateway": "192.168.1.1",
                    "metric": "150",
                },
                '{"dhcp4": false, "addresses": ["192.168.1.69/24"], "routes": [{"to": "0.0.0.0/0", "via": "192.168.1.1", "metric": 150}]}',
            ),
            (
                {
                    "name": "enp0s3",
                    "devId": "pci:0000:00:03.00",
                    "driver": "e1000",
                    "MAC": "08:00:27:2f:be:47",
                    "IPv4": "192.168.1.69/24",
                    "IPv4Mask": "24",
                    "dhcp": "no",
                    "gateway": "192.168.1.1",
                    "metric": "",
                },
                '{"dhcp4": false, "addresses": ["192.168.1.69/24"], "routes": [{"to": "0.0.0.0/0", "via": "192.168.1.1", "metric": 0}]}',
            ),
            (
                {
                    "name": "enp0s3",
                    "devId": "pci:0000:00:03.00",
                    "driver": "e1000",
                    "MAC": "08:00:27:2f:be:47",
                    "IPv4": "192.168.1.69/24",
                    "IPv4Mask": "24",
                    "dhcp": "yes",
                    "gateway": "192.168.1.1",
                    "metric": "150",
                    "dnsDomains": ["yandex.com", "yandex.ru"],
                    "dnsServers": ["8.8.8.8", "1.1.1.1"],
                },
                '{"dhcp4": true, "dhcp4-overrides": {"route-metric": 150, "use-dns": true}, "nameservers": {"addresses": ["8.8.8.8", "1.1.1.1"], "search": ["yandex.com", "yandex.ru"]}}',
            ),
            (
                {
                    "name": "enp0s3",
                    "devId": "pci:0000:00:03.00",
                    "driver": "e1000",
                    "MAC": "08:00:27:2f:be:47",
                    "IPv4": "192.168.1.69/24",
                    "IPv4Mask": "24",
                    "dhcp": "yes",
                    "gateway": "192.168.1.1",
                    "metric": "150",
                    "useDhcpDnsServers": False,
                },
                '{"dhcp4": true, "dhcp4-overrides": {"route-metric": 150, "use-dns": false}}',
            ),
            (
                {
                    "name": "enp0s3",
                    "devId": "pci:0000:00:03.00",
                    "driver": "e1000",
                    "MAC": "08:00:27:2f:be:47",
                    "IPv4": "192.168.1.69/24",
                    "IPv4Mask": "24",
                    "dhcp": "yes",
                    "gateway": "192.168.1.1",
                    "metric": "150",
                    "useDhcpDnsServers": False,
                    "dnsDomains": ["yandex.com", "yandex.ru"],
                    "dnsServers": ["8.8.8.8", "1.1.1.1"],
                },
                '{"dhcp4": true, "dhcp4-overrides": {"route-metric": 150, "use-dns": false}, "nameservers": {"addresses": ["8.8.8.8", "1.1.1.1"], "search": ["yandex.com", "yandex.ru"]}}',
            ),
            (
                {
                    "name": "enp0s3",
                    "devId": "pci:0000:00:03.00",
                    "driver": "e1000",
                    "MAC": "08:00:27:2f:be:47",
                    "IPv4": "192.168.1.69/24",
                    "IPv4Mask": "24",
                    "dhcp": "yes",
                    "gateway": "192.168.1.1",
                    "metric": "150",
                    "useDhcpDnsServers": True,
                    "dnsDomains": ["yandex.com", "yandex.ru"],
                    "dnsServers": ["8.8.8.8", "1.1.1.1"],
                },
                '{"dhcp4": true, "dhcp4-overrides": {"route-metric": 150, "use-dns": true}, "nameservers": {"addresses": ["8.8.8.8", "1.1.1.1"], "search": ["yandex.com", "yandex.ru"]}}',
            ),
            (
                {
                    "name": "enp0s3",
                    "devId": "pci:0000:00:03.00",
                    "driver": "e1000",
                    "MAC": "08:00:27:2f:be:47",
                    "IPv4": "192.168.1.69/24",
                    "IPv4Mask": "24",
                    "dhcp": "yes",
                    "gateway": "192.168.1.1",
                    "metric": "",
                    "dnsDomains": ["yandex.com", "yandex.ru"],
                    "dnsServers": ["8.8.8.8", "1.1.1.1"],
                },
                '{"dhcp4": true, "dhcp4-overrides": {"route-metric": 0, "use-dns": true}, "nameservers": {"addresses": ["8.8.8.8", "1.1.1.1"], "search": ["yandex.com", "yandex.ru"]}}',
            ),
            (
                {
                    "name": "enp0s3",
                    "devId": "pci:0000:00:03.00",
                    "driver": "e1000",
                    "MAC": "08:00:27:2f:be:47",
                    "IPv4": "192.168.1.69/24",
                    "IPv4Mask": "24",
                    "dhcp": "yes",
                    "gateway": "192.168.1.1",
                    "dnsDomains": ["yandex.com", "yandex.ru"],
                    "dnsServers": ["8.8.8.8", "1.1.1.1"],
                },
                '{"dhcp4": true, "dhcp4-overrides": {"route-metric": 0, "use-dns": true}, "nameservers": {"addresses": ["8.8.8.8", "1.1.1.1"], "search": ["yandex.com", "yandex.ru"]}}',
            ),
        ],
    )
    def test_generate_netplan_configuration_success(self, interface, expected):
        self.my_setup()

        # current_dir = os.path.dirname(os.path.realpath(__file__))
        # output_dir = os.path.join(current_dir, "output")

        mediator = Mediator()
        result = mediator.generate_netplan_configuration(interface)

        print("generate_netplan_configuration returned:")
        print(json.dumps(result))

        assert json.dumps(result) == expected

    @pytest.mark.parametrize(
        "files_to_create,expected",
        [
            (
                [],
                "01-netcfg.baseline.yaml",
            ),
            (
                ["01-netcfg.yaml"],
                "02-netcfg.baseline.yaml",
            ),
            (
                ["01-netcfg.yaml", "02-netcfg.yaml", "03-netcfg.yaml"],
                "04-netcfg.baseline.yaml",
            ),
            (
                ["01-netcfg.yaml", "03-netcfg.yaml", "04-netcfg.yaml"],
                "02-netcfg.baseline.yaml",
            ),
            (
                ["network.yaml"],
                "01-netcfg.baseline.yaml",
            ),
            (
                ["01-netcfg.baseline.yaml"],
                "02-netcfg.baseline.yaml",
            ),
            (
                ["01-netcfg.yaml", "02-netcfg.baseline.yaml"],
                "03-netcfg.baseline.yaml",
            ),
        ],
    )
    def test_get_next_available_netplan_file_name_success(
        self, files_to_create, expected
    ):
        self.my_setup()

        current_dir = os.path.dirname(os.path.realpath(__file__))
        output_dir = os.path.join(current_dir, "output")

        # generate some files in the directory
        for file_to_create in files_to_create:
            os.mknod(os.path.join(output_dir, file_to_create))

        mediator = Mediator()
        result = mediator.get_next_available_netplan_file_name(output_dir)

        assert result == expected

    @pytest.mark.parametrize(
        "interface_name,file_name,expected",
        [
            ("enp3s0", "01-dhcp-overrides.yaml", True),
            ("enp3s0", "02-dhcp-overrides.yaml", True),
            ("enp3s0", "03-dhcp-overrides.yaml", True),
            ("enp3s0", "04-dhcp-overrides.yaml", False),
            ("unknown", "04-dhcp-overrides.yaml", True),
            ("enp3s0", "05-dhcp-overrides.yaml", True),
        ],
    )
    def test_is_dns_from_dhcp(self, interface_name, file_name, expected):
        self.my_setup()

        current_dir = os.path.dirname(os.path.realpath(__file__))
        input_dir = os.path.join(current_dir, "input/dhcp-overrides")
        input_file = os.path.join(input_dir, file_name)
        # input_file = os.path.join(current_dir, file_name)

        result = _is_dns_from_dhcp(interface_name, input_file)

        assert result == expected

    @pytest.mark.skip(
        reason="since using soft_check_multiple_interface_definitions with hard-coded paths, no way to test right now"
    )
    def test_update_device_interfaces_failure(self):
        self.my_setup()

        current_dir = os.path.dirname(os.path.realpath(__file__))
        output_dir = os.path.join(current_dir, "output")

        # both 01-netcfg.yaml and 02-netcfg.yaml have same enp0s3 interface configured

        file_path = os.path.join(current_dir, "input/01-netcfg.yaml")
        cmd = "cp %s %s" % (file_path, output_dir)
        os.system(cmd)

        file_path = os.path.join(current_dir, "input/02-netcfg.yaml")
        cmd = "cp %s %s" % (file_path, output_dir)
        os.system(cmd)

        # netplan_dirs = lambda: [output_dir]

        mediator = Mediator()
        result = mediator.update_device_interfaces(
            {}, get_netplan_dirs=_get_netplan_dirs
        )

        assert True == False
